#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "clock.h"
#include "wav.h"

#ifdef DEBUG
#define DACTION( act ) act
#else
#define DACTION( act )
#endif

/* Filters */
void lowpass( wav_file *wav ) {
  uint64_t clockStart = x86clock();
  const float coefficients[4] = { 0.378f, 0.623f, 0.623f, 0.378f };
  int32_t __attribute__((aligned(16))) samples[4] = {
    wav->samples[0], wav->samples[0], wav->samples[1], wav->samples[2] 
  };
  
#ifdef USE_SSE
  float result[4];
#endif

  int i;
  for( i = 3; i < wav->numsamples; i++ ) {

    /* Neuen Wert einlesen */
    samples[0] = samples[1];
    samples[1] = samples[2];
    samples[2] = samples[3];
    samples[3] = wav->samples[i];

    DACTION( printf( "%f * %d + %f * %d + %f * %d + %f * %d = ",
       coefficients[3], wav->samples[i - 3],
       coefficients[2], wav->samples[i - 2],
       coefficients[1], wav->samples[i - 1],
       coefficients[0], wav->samples[i - 0] ) );
#ifdef USE_SSE
    /*
     * Verwendete Befehle
     * ctdq2ps - Convert Packed Signed Doubleword Integers to Packed Single-Precision Floating-Point Values
     * movups - Move Unaligned Packed Single-Precision Floating- Point Values
     * mulps - Multiply Packed Single-Precision Floating-Point Values
     * haddps - Packed Single-FP Horizontal Add
     */
    asm(
        "cvtdq2ps %[samples], %%xmm0\n"
        "movups %[coefs], %%xmm1\n"
        "mulps  %%xmm1, %%xmm0\n"
        "haddps %%xmm0, %%xmm0\n"
        "haddps %%xmm0, %%xmm0\n"
        "movups %%xmm0, %[result]\n"
        : /* Output */
        [result] "=m"( result[0] )
        : /* Input */
        [samples] "m"( samples[0] ),
        [coefs] "m"( coefficients[0] )
        : /* Clobbers */
        "xmm0", "xmm1" );
    
    wav->samples[i] = result[0];
#else
    float accu = 0;
    int j;
    for ( j = 0; j < 4; j++ ) {
      accu += samples[j] * coefficients[j];
    }
    wav->samples[i] = accu;
#endif
    DACTION( printf( " %d\n", wav->samples[i] ) );
  }
  
  uint64_t clockEnd = x86clock();
  printf( "Time taken: %lli\n", clockEnd - clockStart );
}

void noise( wav_file *wav, int thresh ) {
  int i, a, b;
  for( i = 0; i < wav->numsamples; i++ ) {
    if( abs( wav->samples[i] ) < thresh ) {
      a = rand() % thresh;
      b = rand() % 1024;
      if( b & 1 )
        a = -a;
      wav->samples[i] += a;
    }
  }
}

void echo( wav_file *wav, unsigned int Delay, unsigned int NumEcho ) {
  uint64_t clockStart = x86clock();
  
  int i, k;
  
  // Start echoing
#if defined USE_MMX || defined USE_MMX_2
  for( k = 1; k <= NumEcho; k++ ) {
#ifdef USE_MMX
    const int unroll_factor = 4;
#else
    const int unroll_factor = 8;
#endif

    // Execute first few iterations until iteration modulo 4 is possible
    for( i = wav->numsamples - 1; 
         i >= Delay*k && ( ( i - Delay*k ) % unroll_factor != 0 );
         i-- ) {
      DACTION( int16_t old_value = wav->samples[i] );
      wav->samples[i] += wav->samples[i - Delay * k] >> k;
      DACTION( printf( "%10d: %d + (%d >> %d) = %d\n", i,
        old_value, wav->samples[i - Delay * k], k, wav->samples[i] ) );
    }

    // Execute unrolled iterations
    for( ; i > Delay*k; i -= unroll_factor ) {
      DACTION( int16_t old_values[4];
               old_values[0] = wav->samples[i - 0];
               old_values[1] = wav->samples[i - 1];
               old_values[2] = wav->samples[i - 2];
               old_values[3] = wav->samples[i - 3]; );
      /*
       * Entrollen um Faktor 4 zeigt Parallelität
       *
      wav->samples[i    ] += wav->samples[i     - Delay * k] >> k;
      wav->samples[i - 1] += wav->samples[i - 1 - Delay * k] >> k;
      wav->samples[i - 2] += wav->samples[i - 2 - Delay * k] >> k;
      wav->samples[i - 3] += wav->samples[i - 3 - Delay * k] >> k;
      */

#ifdef USE_MMX
      /*
       * Verwendete Befehle
       * movq  - Move Quadword
       * moqd  - Move Doubleword
       * psraw - Shift Packed Data Right Arithmetic
       * paddw - Add Packed Integers
       */
      asm(
          "movq %[samples_in], %%mm0\n"
          "movq %[echos], %%mm1\n"
          "movd %[shift], %%mm2\n"
          "psraw %%mm2, %%mm1\n"
          "paddw %%mm1, %%mm0\n"
          "movq %%mm0, %[samples_out]\n"
          : // Output
          [samples_out] "=m"( wav->samples[i - unroll_factor + 1] )
          : // Input
          [samples_in] "m"( wav->samples[i - unroll_factor + 1] ),
          [echos] "m"( wav->samples[i - unroll_factor + 1 - Delay * k] ),
          [shift] "r"( k )
          : /* Clobber */
          "mm0", "mm1", "mm2" );
#else
#ifdef USE_SSE
    /* Besser: (SSE-Parallelität nutzen) */
      asm(
          "movups %[samples_in], %%xmm0\n"
          "movups %[echos], %%xmm1\n"
          "movd %[shift], %%xmm2\n"
          "psraw %%xmm2, %%xmm1\n"
          "paddw %%xmm1, %%xmm0\n"
          "movups %%xmm0, %[samples_out]\n"
          : // Output
          [samples_out] "=m"( wav->samples[i - unroll_factor + 1] )
          : // Input
          [samples_in] "m"( wav->samples[i - unroll_factor + 1] ),
          [echos] "m"( wav->samples[i - unroll_factor + 1 - Delay * k] ),
          [shift] "r"( k )
          : // Clobber
          "xmm0", "xmm1", "xmm2" );
#else
  for( k = 1; k <= NumEcho; k++ ) {
    for( i = wav->numsamples - 1; i >= Delay*k; i-- ) {
      wav->samples[i] += wav->samples[i - Delay * k] >> k;
    }
  }
#endif
#endif

      DACTION( int j;
      for ( j = 0; j < unroll_factor; j++ ) {
        printf( "%10d: %d + (%d >> %d) = %d\n", i -j, 
          old_values[j], wav->samples[i - j - Delay * k], k, 
          wav->samples[i - j] );
      } );
    }
  }

  asm( "emms\n" );
#else
  for( k = 1; k <= NumEcho; k++ ) {
    for( i = wav->numsamples - 1; i >= Delay*k; i-- ) {
      DACTION( int16_t old_value = wav->samples[i] );
      wav->samples[i] += wav->samples[i - Delay * k] >> k;
      DACTION( printf( "%10d: %d + (%d >> %d) = %d\n", i,
        old_value, wav->samples[i - Delay * k], k, wav->samples[i] ) );
    }
  }
#endif
  
  uint64_t clockEnd = x86clock();
  printf( "Time taken: %lli\n", clockEnd - clockStart );
}

/* Main Program */
int main(int argc,char **argv)
{
  // Check arguments
  if (argc<3) {
    printf("Usage: %s <input-wav filename> <filter-name> <output-wav filename>\n",argv[0]);
    return(1);
  }
  char *inputFile = argv[1];
  char *outputFile = argv[3];
  char *filter = argv[2];

  // Read WAV  
  wav_file *inputWAV = readWAV( inputFile );

  // Filter
  if ( strcmp( filter, "echo" ) == 0 ) {
    echo( inputWAV, 500, 10 );
  } else
  if ( strcmp( filter, "lowpass" ) == 0 ) {
    lowpass( inputWAV );
  } else
  if ( strcmp( filter, "noise" ) == 0 ) {
    noise( inputWAV, 750 );
  } else {
    printf("Unknown filter '%s'!\n", filter);
    exit(3);
  }

  // Write result
  writeWAV( inputWAV, outputFile );

  // Free the WAV struct
  freeWAV( inputWAV );

  return 0;
}
