#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "wav.h"

wav_file *readWAV( const char *inputFile ) {
  wav_file *result = ( wav_file* )malloc( sizeof( wav_file ) );
  FILE *f;
  if( ( f = fopen( inputFile, "rb" ) ) ) {
    /* Lese Header */
    if ( fread( &result->OrigHeader, 1, sizeof(result->OrigHeader), f ) < sizeof(result->OrigHeader) ) {
      printf("fread() failed for %s\n", inputFile);
      fclose(f);
      exit(1);
    }
    if( ( strncmp( result->OrigHeader.id, "RIFF", 4 ) ) || 
        ( strncmp( result->OrigHeader.type, "WAVE", 4 ) ) ) {
      printf( "error, the provided file is not a valid .wav file\n" );
      fclose( f );
      exit( 1 );
    }
    
    /* Lese und prüfe Format */
    fread( &result->OrigFormat, 1, 24, f );
    printf( "# sample format: %u\nchannels: %u\nsample rate: %u Hz\nbytes per sample: %u\nbits per sample: %u\n",
				    result->OrigFormat.compcode,
            result->OrigFormat.channels,
            result->OrigFormat.samplerate,
            result->OrigFormat.bytespersample,
            result->OrigFormat.bitspersample );
            
    if( ( result->OrigFormat.bitspersample != 16 ) ||
        ( result->OrigFormat.channels != 1 ) ) {
      printf( "error, only capable of handling 16 bit mono sample data\n" );
      fclose( f );
      exit( 1 );
    }
    
    /* Lese Anzahl Samples */
    fread( &result->OrigData, 1, 8, f );
    result->numsamples = ( ( result->OrigData.length ) >> 1 );
      
    /* Speicher für Samples allokieren und nullen */
    result->samples = ( short * )malloc( result->numsamples * sizeof( short ) );
    memset( result->samples, 0, result->numsamples * sizeof( short ) );
    
    /* Samples einlesen */
    fread( result->samples, 1, result->OrigData.length, f );

    /* Datei schließen */
    fclose( f );
  } else {
  printf( "error, could not open file %s\n", inputFile );
    exit( 1 );
  }
  
  return result;
}

void writeWAV( const wav_file *wav, const char *outputFile ) 
{
  FILE *f;
  if ( ( f = fopen( outputFile, "wb" ) ) ) {
    fwrite(&wav->OrigHeader,1,12,f);
    fwrite(&wav->OrigFormat,1,24,f);
    fwrite(&wav->OrigData,1,8,f);
    fwrite(&wav->samples,1,wav->OrigData.length,f);
    fclose(f);
  } else {
    printf("error, could not open file %s for writing\n",outputFile);
    exit(2);
  }
}

void freeWAV( wav_file *wav )
{
  free( wav->samples );
  free( wav );
}
